/*
 *  KANJI Code conversion routines.
 */

#include <ptxconfig.h>
#include <kanji.h>
#include <kanjicnv.h>
#include <unicode-jp.h>
#include <stdio.h>  /* sprintf() */
#include <stdlib.h>

#ifdef DEFAULT_KANJI

#define JIS   0
#define EUC   1
#define SJIS  2
#define UTF8  3

#define ESC '\033'

#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#else
#include <limits.h>
#endif /* HAVE_SYS_PARAM_H */
#ifndef OPEN_MAX
#define OPEN_MAX 32
#endif
#ifndef NOFILE
#define NOFILE OPEN_MAX
#endif

#ifdef WIN32
PTXDLL int sjisterminal = 0;
#endif

static int kanji_enc = DEFAULT_KANJI;
static int internal_euc_only = false;

int get_kanji_enc (void)
{
    return kanji_enc;
}

void set_kanji_enc (int encode)
{
    kanji_enc = encode;
}

char * get_kanji_enc_string (void)
{
    switch (kanji_enc) {
    case JIS:  return "jis";
    case EUC:  return "euc";
    case SJIS: return "sjis";
    case UTF8: return "utf8";
    default:   return "?";
    }
}

int set_kanji_enc_string (char * encode)
{
    if (strcasecmp(encode, "jis") == 0) {
        kanji_enc = JIS;
    } else if (strcasecmp(encode, "euc") == 0) {
        kanji_enc = EUC;
    } else if (strcasecmp(encode, "sjis") == 0) {
        kanji_enc = SJIS;
    } else if (strcasecmp(encode, "utf8") == 0) {
        kanji_enc = UTF8;
    } else {
        return -1;
    }
    return kanji_enc;
}

void set_internal_euc_only (void)
{
    internal_euc_only = true;
}

int is_internalSJIS (void)
{
    if (internal_euc_only) return false;
    return (kanji_enc == SJIS);
}


int iskanji1 (int c)
{
    if (is_internalSJIS()) return isSJISkanji1(c);
    else                   return isEUCkanji1(c);
}

int iskanji2 (int c)
{
    if (is_internalSJIS()) return isSJISkanji2(c);
    else                   return isEUCkanji2(c);
}

/* JIS to EUC/SJIS Kanji code conversion */
int toJIS (int kcode)
{
    if (is_internalSJIS()) return SJIStoJIS(kcode);
    else                   return EUCtoJIS(kcode);
}

/* JIS to EUC/SJIS Kanji code conversion */
int fromJIS (int kcode)
{
    if (is_internalSJIS()) return JIStoSJIS(kcode);
    else                   return JIStoEUC(kcode);
}

/* EUC to EUC/SJIS Kanji code conversion */
int fromEUC (int kcode)
{
    if (is_internalSJIS()) return EUCtoSJIS(kcode);
    else                   return kcode;
}

/* SJIS to EUC/SJIS Kanji code conversion */
int fromSJIS (int kcode)
{
    if (is_internalSJIS()) return kcode;
    else                   return SJIStoEUC(kcode);
}

/* KUTEN to EUC/SJIS Kanji code conversion */
int fromKUTEN (int kcode)
{
    return fromJIS(KUTENtoJIS(kcode));
}


/* put char with Kanji code conversion */
int putc2 (int c, FILE* fp)
{
    static int inkanji[NOFILE]; /* 0: not in Kanji
                                   1: in JIS Kanji and first byte is in c1[]
                                   2: in JIS Kanji */
    static unsigned char c1[NOFILE];
    const int fd = fileno(fp);
    int i, ret = c;

    if (inkanji[fd] == 1) {
        i = HILO(c1[fd], c);
        if (kanji_enc == UTF8) {
            if(sjisterminal && (fd == 1 || fd == 2)) {
                i = JIStoSJIS(toJIS(i));
                ret = putc(HI(i), fp);
                ret = putc(LO(i), fp);
	    } else {
                i = JIStoUCS2(toJIS(i));
                if (i < 0x800) {
                    ret = putc(UCS2toUTF8B1(i), fp);
                    ret = putc(UCS2toUTF8B2(i), fp);
                } else {
                    ret = putc(UCS2toUTF8C1(i), fp);
                    ret = putc(UCS2toUTF8C2(i), fp);
                    ret = putc(UCS2toUTF8C3(i), fp);
                }
	    }
        } else {
            if(sjisterminal && (fd == 1 || fd == 2)) {
                if(internal_euc_only || kanji_enc != SJIS) i = EUCtoSJIS(i);
	    } else {
                if(internal_euc_only && kanji_enc == SJIS) i = EUCtoSJIS(i);
                else if (kanji_enc == JIS) i = EUCtoJIS(i);
	    }
            ret = putc(HI(i), fp);
            ret = putc(LO(i), fp);
        }
        inkanji[fd] = 2;
    } else if (iskanji1(c)) {
        if (kanji_enc == JIS && inkanji[fd] == 0) {
            if(sjisterminal && (fd == 1 || fd == 2)) {
                ;
	    } else {
                ret = putc(ESC, fp);
                ret = putc('$', fp);
                ret = putc('B', fp); /* kanji in */
	    }
        }
        inkanji[fd] = 1;
        c1[fd] = c; /* store 'c' and do not put this time */
    } else {
        if (kanji_enc == JIS && inkanji[fd] == 2) {
            if(sjisterminal && (fd == 1 || fd == 2)) {
                ;
	    } else {
                ret = putc(ESC, fp);
                ret = putc('(', fp);
                ret = putc('B', fp); /* kanji out */
	    }
        }
        inkanji[fd] = 0;
        ret = putc(c, fp);
    }
    return ret;
}

static char * buffer;
static long first, last, bufsize;
static int combin_voiced_sound (int semi)
{
    int i;

    if (last-2 < first) return false;
    if (!iskanji1(buffer[last-2])) return false;
    if (!iskanji2(buffer[last-1])) return false;
    i = HILO(buffer[last-2], buffer[last-1]);
    i = JIStoUCS2(toJIS(i));
    i = get_voiced_sound(i, semi);
    if (i == 0) return false;
    i = fromJIS(UCS2toJIS(i));
    buffer[last-2] = HI(i);
    buffer[last-1] = LO(i);
    return true;
}

static void write_hex (int i)
{
    if (last + 4 < bufsize) {
        sprintf(buffer + last, "^^%02x", i);
        last += 4;
    } else {
        last = bufsize;
    }
}

void get_utf8 (int i, FILE *f)
{
    int j, u, i2=-1, i3=-1, i4=-1;

    if (i < 0xe0) {
        i2 = getc(f);
        u = UTF8BtoUCS2(i, i2);
        j = UCS2toJIS(u);
    } else if (i < 0xf0) {
        i2 = getc(f); i3 = getc(f);
        u = UTF8CtoUCS2(i, i2, i3);
        if (u == U_BOM) return; /* just ignore */
        if (u == U_VOICED      && combin_voiced_sound(false)) return;
        if (u == U_SEMI_VOICED && combin_voiced_sound(true))  return;
        j = UCS2toJIS(u);
    } else if (i < 0xf8) {
        i2 = getc(f); i3 = getc(f); i4 = getc(f);
        j = 0; /* unsupported */
    } else {
        j = 0; /* broken */
    }

    if (j == 0) {
        write_hex(i);
        if (i2 >= 0) write_hex(i2);
        if (i3 >= 0) write_hex(i3);
        if (i4 >= 0) write_hex(i4);
    } else {
        i = fromJIS(j);
        buffer[last++] = HI(i);
        buffer[last++] = LO(i);
    }
}

void get_sjis (int i, FILE *f)
{
    int j;

    j = getc(f);
    if (isSJISkanji2(j)) {
        i = SJIStoEUC(HILO(i,j));
        buffer[last++] = HI(i);
        buffer[last++] = LO(i);
    } else {
        ungetc(j, f);
        buffer[last++] = i;
    }
}        

int is_tail (int *c, FILE *fp)
{
    if (*c == EOF) return true;
    if (*c == '\n') return true;
    if (*c == '\r') {
        int d = getc(fp);
        if (d == '\n') *c = d;
        else ungetc(d, fp);
        return true;
    }
    return false;
}


/* input line with Kanji code conversion */
int input_line2 (FILE *fp, char * buff, long pos, long size,
                    int *lastchar)
{
    int i;
    static int injis = false;

    buffer = buff;
    first = last = pos;
    bufsize = size;

    while (last < bufsize-30 && (i=getc(fp)) != EOF && i!='\n' && i!='\r') {
        if (i == ESC) {
            if ((i=getc(fp)) == '$') { /* ESC '$' (Kanji-in) */
                i = getc(fp);
                if (i == '@' || i == 'B') {
                    injis = true;
                } else {               /* broken Kanji-in */
                    buffer[last++] = ESC;
                    buffer[last++] = '$';
                    if (is_tail(&i, fp)) break;
                    buffer[last++] = i;
                }
            } else if (i == '(') {     /* ESC '(' (Kanji-out) */
                i = getc(fp);
                if (i == 'J' || i == 'B' || i == 'H') {
                    injis = false;
                } else {               /* broken Kanji-out */
                    buffer[last++] = ESC;
                    buffer[last++] = '(';
                    if (is_tail(&i, fp)) break;
                    buffer[last++] = i;
                }
            } else { /* broken ESC */
                buffer[last++] = ESC;
                if (is_tail(&i, fp)) break;
                buffer[last++] = i;
            }
        } else { /* rather than ESC */
            if (injis) { /* in JIS */
                int j = getc(fp);
                if (is_tail(&j, fp)) {
                    buffer[last++] = i;
                    i = j;
                    break;
                } else { /* JIS encoding */
                    i = fromJIS(HILO(i,j));
                    buffer[last++] = HI(i);
                    buffer[last++] = LO(i);
                }
            } else {  /* normal */
                if (internal_euc_only && kanji_enc==SJIS && isSJISkanji1(i)) {
                    get_sjis(i, fp);
                } else if (kanji_enc == UTF8 && i >= 0xc0) {
                    get_utf8(i, fp);
                } else {
                    buffer[last++] = i;
                }
            }
        }
    }

    buffer[last] = '\0';
    if (i == EOF || i == '\n' || i == '\r') injis = false;
    if (lastchar != NULL) *lastchar = i;
    return last;
}

#endif /* DEFAULT_KANJI */
